library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity ram is
   generic
	(
		address_width_ram 	: integer := 8;
		data_width_ram 		: integer := 8
	);
	
	port
	(
		clock_read_ram			: in  std_logic;
		clock_write_ram      : in  std_logic;
		data_in_ram 			: in  std_logic_vector(data_width_ram - 1 downto 0);
		data_out_ram			: out std_logic_vector(data_width_ram - 1 downto 0);
		write_enable_ram  	: in  std_logic;	
		write_address_ram 	: in  std_logic_vector(address_width_ram - 1 downto 0);	
		read_enable_ram      : in 	std_logic;				
		read_address_ram  	: in  std_logic_vector(address_width_ram - 1 downto 0)
	);
end ram;

architecture ram_rtl of ram is
	type ram_type is array(2 ** address_width_ram - 1 downto 0) of std_logic_vector(data_width_ram - 1 downto 0);
	signal ram : ram_type;
begin

	-- This is a dual port ram
	process (clock_read_ram)
	begin
		if rising_edge(clock_read_ram)
		then			
				if read_enable_ram = '1'
				then
					-- Read from RAM
					data_out_ram <= ram(to_integer(unsigned(read_address_ram)));
				else
					data_out_ram <= (others => 'Z');
				end if;		
		end if;
	end process;
	
	process (clock_write_ram)
	begin
		if rising_edge(clock_write_ram)
		then
				if write_enable_ram = '1'
				then
					-- Write to the RAM
					ram(to_integer(unsigned(write_address_ram))) <= data_in_ram;
				else
					null;
				end if;
		end if;
	end process;
end architecture ram_rtl;

-------------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-- This is the extended version of the RAM
-- The RAM is base for a Content Access RAM a CAM
-- This CAM is useable for reading, writing and also searching a content
-- If the search or lookup is successful the address of the content is delivered as output
entity cam is
	generic
	(
		address_width_cam 	: integer := 8;
		data_width_cam 		: integer := 8
	);
	
	port
	(
		------------------------------------ Inputs
		clock_cam				: in  std_logic;
		data_in_cam 			: in  std_logic_vector(data_width_cam - 1 downto 0);		
		write_enable_cam  	: in  std_logic;	
		write_address_cam 	: in  std_logic_vector(address_width_cam - 1 downto 0);	
		read_enable_cam      : in 	std_logic;				
		read_address_cam  	: in  std_logic_vector(address_width_cam - 1 downto 0);
		lookup_cam				: in  std_logic;
		------------------------------------ Outputs
		data_out_cam			: out std_logic_vector(data_width_cam - 1 downto 0);
		address_out_cam      : out std_logic_vector(address_width_cam - 1 downto 0);
		error_cam            : out std_logic
	);
end cam;

architecture cam_rtl of cam is
	component ram is
	generic
	(
		address_width_ram 	: integer := 8;
		data_width_ram 		: integer := 8
	);
	
	port
	(
		--------------------- Inputs
		clock_read_ram			: in  std_logic;
		clock_write_ram      : in  std_logic;
		data_in_ram 			: in  std_logic_vector(data_width_ram - 1 downto 0);	
		write_enable_ram  	: in  std_logic;	
		write_address_ram 	: in  std_logic_vector(address_width_ram - 1 downto 0);	
		read_enable_ram      : in 	std_logic;				
		read_address_ram  	: in  std_logic_vector(address_width_ram - 1 downto 0);
		--------------------- Outputs
		data_out_ram			: out std_logic_vector(data_width_ram - 1 downto 0)
	);
	end component ram;
	
	-- The CAM is modeled as a state machine
	type state is (idle_state, lookup_state, search_state, found_state, not_found_state);
	signal state_cam : state := idle_state;
	
	----------------------------> This are the internal signals for the ports
	signal data_out_cam_internal			: std_logic_vector(data_width_cam - 1 downto 0);
	signal read_address_cam_internal		: std_logic_vector(address_width_cam - 1 downto 0);
	----------------------------> This are additional signal
	signal read_address              	: natural range 2 ** address_width_cam - 1 downto 0;
begin
	generate_ram : if true
	generate
		generate_ram_block:
			ram generic map (address_width_ram 	=> address_width_cam,
								  data_width_ram    	=> data_width_cam)
				 port map (
								--------------- Inputs
								clock_read_ram 		=> clock_cam,
								clock_write_ram      => clock_cam,
								data_in_ram 			=> data_in_cam,							
								write_enable_ram  	=> write_enable_cam,
								write_address_ram 	=> write_address_cam,	
								read_enable_ram      => read_enable_cam,
								read_address_ram  	=> read_address_cam_internal,
								--------------- Outputs
								data_out_ram			=> data_out_cam_internal
							);
	end generate generate_ram;
	
	state_machine : process (clock_cam)
		variable searched_content					: std_logic_vector(0 to data_width_cam - 1);
	begin
		if rising_edge(clock_cam)
		then
			case state_cam is
				when idle_state =>				if read_enable_cam = '1'
														then
															read_address_cam_internal  <= read_address_cam;
															data_out_cam <= data_out_cam_internal;
														elsif write_enable_cam = '1'
														then
															null;
														elsif lookup_cam = '1'
														then
															state_cam <= lookup_state;
														end if;														
				when lookup_state =>				-- Initialize search
														searched_content           := data_in_cam;
														-- Initialize read address
														read_address_cam_internal <= (others => '0');
														-- Convert to integer
														read_address <= to_integer(unsigned(read_address_cam_internal));
														-- Go searching
														state_cam <= search_state;
				when search_state => 			if data_out_cam_internal = searched_content
														then
															state_cam <= found_state;
														elsif read_address < 2 ** address_width_cam - 1
														then
															-- Increment read address
															read_address <= read_address + 1;
															-- Convert to std_logic_vector
															read_address_cam_internal <= std_logic_vector(to_unsigned(read_address,address_width_cam));
														else
															state_cam <= not_found_state;
														end if;									
				when not_found_state => 		error_cam			<= '1';
														address_out_cam    <= (others => 'Z');
				when found_state =>				error_cam 			<= '0';
														address_out_cam 				<= read_address_cam_internal;				
			end case;
		end if;
	end process state_machine;
end architecture cam_rtl;